from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import Any

from omu.api import Extension, ExtensionType
from omu.api.endpoint import EndpointType
from omu.event_emitter import EventEmitter, Unlisten
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.omu import Omu
from omu.serializer import SerializeError, Serializer

from .packets import RegistryPacket, RegistryRegisterPacket
from .registry import Registry, RegistryType

REGISTRY_EXTENSION_TYPE = ExtensionType("registry", lambda client: RegistryExtension(client))

REGISTRY_PERMISSION_ID = REGISTRY_EXTENSION_TYPE / "permission"

REGISTRY_REGISTER_PACKET = PacketType[RegistryRegisterPacket].create_serialized(
    REGISTRY_EXTENSION_TYPE,
    "register",
    serializer=RegistryRegisterPacket,
)
REGISTRY_UPDATE_PACKET = PacketType[RegistryPacket].create_serialized(
    REGISTRY_EXTENSION_TYPE,
    "update",
    serializer=RegistryPacket,
)
REGISTRY_LISTEN_PACKET = PacketType[Identifier].create_json(
    REGISTRY_EXTENSION_TYPE,
    "listen",
    Serializer.model(Identifier),
)
REGISTRY_GET_ENDPOINT = EndpointType[Identifier, RegistryPacket].create_serialized(
    REGISTRY_EXTENSION_TYPE,
    "get",
    request_serializer=Serializer.model(Identifier).to_json(),
    response_serializer=RegistryPacket,
    permission_id=REGISTRY_PERMISSION_ID,
)


class RegistryExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return REGISTRY_EXTENSION_TYPE

    def __init__(self, omu: Omu) -> None:
        self.client = omu
        self.registries: dict[Identifier, Registry] = {}
        omu.network.register_packet(
            REGISTRY_REGISTER_PACKET,
            REGISTRY_LISTEN_PACKET,
            REGISTRY_UPDATE_PACKET,
        )

    def create_registry[T](self, registry_type: RegistryType[T]) -> Registry[T]:
        self.client.permissions.require(REGISTRY_PERMISSION_ID)
        if registry_type.id in self.registries:
            raise ValueError(f"Registry {registry_type.id} already exists")
        return RegistryImpl(
            self.client,
            registry_type,
        )

    def get[T](self, registry_type: RegistryType[T]) -> Registry[T]:
        return self.create_registry(registry_type)

    def create[T](self, name: str, *, default: T) -> Registry[T]:
        identifier = self.client.app.id / name
        registry_type = RegistryType(
            identifier,
            default,
            Serializer.json(),
        )
        return self.create_registry(registry_type)


class RegistryImpl[T](Registry[T]):
    def __init__(
        self,
        omu: Omu,
        registry_type: RegistryType[T],
    ) -> None:
        self.client = omu
        self.type = registry_type
        self._value = registry_type.default_value
        self.event_emitter: EventEmitter[T] = EventEmitter()
        self.listening = False
        omu.network.add_packet_handler(REGISTRY_UPDATE_PACKET, self._handle_update)
        omu.network.add_task(self._on_ready_task)

    @property
    def value(self) -> T:
        return self._value

    async def get(self) -> T:
        result = await self.client.endpoints.call(REGISTRY_GET_ENDPOINT, self.type.id)
        if result.value is None:
            return self.type.default_value
        try:
            return self.type.serializer.deserialize(result.value)
        except SerializeError as e:
            raise SerializeError(f"Failed to deserialize registry value for identifier {self.type.id}") from e

    async def set(self, value: T) -> None:
        packet = RegistryPacket(
            id=self.type.id,
            value=self.type.serializer.serialize(value),
        )
        await self.client.send(
            REGISTRY_UPDATE_PACKET,
            packet,
        )
        await self.event_emitter.emit(value)

    async def update(self, handler: Coro[[T], T] | Callable[[T], T]) -> T:
        value = await self.get()
        if asyncio.iscoroutinefunction(handler):
            new_value = await handler(value)
        else:
            new_value: T = handler(value)  # type: ignore
        await self.set(new_value)
        return new_value

    async def modify(self, handler: Coro[[T], Any] | Callable[[T], Any]) -> T:
        value = await self.get()
        if asyncio.iscoroutinefunction(handler):
            await handler(value)
        else:
            handler(value)
        await self.set(value)
        return value

    def listen(self, handler: Coro[[T], None] | Callable[[T], None]) -> Unlisten:
        if not self.listening and self.client.ready:
            coro = self.client.send(REGISTRY_LISTEN_PACKET, self.type.id)
            self.client.loop.create_task(coro)
        self.listening = True

        return self.event_emitter.listen(handler)

    async def _handle_update(self, event: RegistryPacket) -> None:
        if event.id != self.type.id:
            return
        if event.value is not None:
            try:
                self._value = self.type.serializer.deserialize(event.value)
            except SerializeError as e:
                msg = f"Failed to deserialize registry value for id {self.type.id}"
                raise SerializeError(msg) from e
        await self.event_emitter.emit(self._value)

    async def _on_ready_task(self) -> None:
        if self.listening:
            await self.client.send(REGISTRY_LISTEN_PACKET, self.type.id)
        if not self.type.id.is_subpath_of(self.client.app.id):
            return
        packet = RegistryRegisterPacket(
            id=self.type.id,
            permissions=self.type.permissions,
        )
        await self.client.send(REGISTRY_REGISTER_PACKET, packet)



from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import Self

from loguru import logger

from omu.helper import Coro

type Unlisten = Callable[[], None]


class EventEmitter[**P]:
    def __init__(
        self,
        on_subscribe: Callable[[], None] | Coro[[], None] | None = None,
        on_empty: Callable[[], None] | Coro[[], None] | None = None,
        catch_errors: bool = False,
    ) -> None:
        self.on_subscribe = on_subscribe
        self.on_empty = on_empty
        self.catch_errors = catch_errors
        self._listeners: list[Callable[P, None] | Coro[P, None]] = []
        self.closed = False

    @property
    def empty(self) -> bool:
        return len(self._listeners) == 0

    def close(self) -> None:
        self.closed = True
        self._listeners.clear()

    def listen(self, listener: Callable[P, None] | Coro[P, None]) -> Unlisten:
        if self.closed:
            raise ValueError("EventEmitter is closed")
        if listener in self._listeners:
            raise ValueError("Listener already subscribed")
        if self.on_subscribe and len(self._listeners) == 0:
            coroutine = self.on_subscribe()
            if asyncio.iscoroutine(coroutine):
                asyncio.create_task(coroutine)
        self._listeners.append(listener)
        return lambda: self.unlisten(listener)

    def unlisten(self, listener: Callable[P, None] | Coro[P, None]) -> None:
        if listener not in self._listeners:
            return
        self._listeners.remove(listener)
        if self.on_empty and len(self._listeners) == 0:
            coroutine = self.on_empty()
            if asyncio.iscoroutine(coroutine):
                asyncio.create_task(coroutine)

    async def emit(self, *args: P.args, **kwargs: P.kwargs) -> None:
        if self.closed:
            raise ValueError("EventEmitter is closed")
        for listener in tuple(self._listeners):
            try:
                if asyncio.iscoroutinefunction(listener):
                    await listener(*args, **kwargs)
                else:
                    listener(*args, **kwargs)
            except Exception as e:
                if self.catch_errors:
                    logger.opt(exception=e).error("Error in listener")
                else:
                    raise e

    def __iadd__(self, listener: Callable[P, None] | Coro[P, None]) -> Self:
        self.listen(listener)
        return self

    __call__ = emit



import abc


class Keyable(abc.ABC):
    @abc.abstractmethod
    def key(self) -> str: ...

    def __hash__(self) -> int:
        return hash(self.key())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.key()})"



from __future__ import annotations

from omu.api import Extension, ExtensionType
from omu.api.endpoint import EndpointType
from omu.api.registry import RegistryPermissions, RegistryType
from omu.api.table import TablePermissions, TableType
from omu.app import App
from omu.omu import Omu

SERVER_EXTENSION_TYPE = ExtensionType("server", lambda client: ServerExtension(client))

SERVER_APPS_READ_PERMISSION_ID = SERVER_EXTENSION_TYPE / "apps" / "read"
SERVER_APP_TABLE_TYPE = TableType.create_model(
    SERVER_EXTENSION_TYPE,
    "apps",
    App,
    permissions=TablePermissions(
        read=SERVER_APPS_READ_PERMISSION_ID,
    ),
)
SERVER_SHUTDOWN_PERMISSION_ID = SERVER_EXTENSION_TYPE / "shutdown"
SHUTDOWN_ENDPOINT_TYPE = EndpointType[bool, bool].create_json(
    SERVER_EXTENSION_TYPE,
    "shutdown",
    permission_id=SERVER_SHUTDOWN_PERMISSION_ID,
)
TRUSTED_ORIGINS_GET_PERMISSION_ID = SERVER_EXTENSION_TYPE / "trusted_origins" / "get"
TRUSTED_ORIGINS_SET_PERMISSION_ID = SERVER_EXTENSION_TYPE / "trusted_origins" / "set"
TRUSTED_ORIGINS_REGISTRY_TYPE = RegistryType[list[str]].create_json(
    SERVER_EXTENSION_TYPE,
    "trusted_origins",
    default_value=[],
    permissions=RegistryPermissions(
        read=TRUSTED_ORIGINS_GET_PERMISSION_ID,
        write=TRUSTED_ORIGINS_SET_PERMISSION_ID,
    ),
)


class ServerExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return SERVER_EXTENSION_TYPE

    def __init__(self, omu: Omu) -> None:
        self._client = omu
        self.apps = omu.tables.get(SERVER_APP_TABLE_TYPE)
        self.sessions = omu.tables.get(SERVER_APP_TABLE_TYPE)
        self.trusted_origins = omu.registries.get(TRUSTED_ORIGINS_REGISTRY_TYPE)

    async def shutdown(self, restart: bool = False) -> bool:
        return await self._client.endpoints.call(SHUTDOWN_ENDPOINT_TYPE, restart)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from omu.api import Extension, ExtensionType
from omu.api.endpoint import EndpointType
from omu.bytebuffer import ByteReader, ByteWriter
from omu.identifier import Identifier
from omu.serializer import Serializer

if TYPE_CHECKING:
    from omu.omu import Omu

ASSET_EXTENSION_TYPE = ExtensionType("asset", lambda client: AssetExtension(client))


@dataclass(frozen=True, slots=True)
class Asset:
    id: Identifier
    buffer: bytes


class FileSerializer:
    @classmethod
    def serialize(cls, item: Asset) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_uint8_array(item.buffer)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> Asset:
        with ByteReader(item) as reader:
            identifier = Identifier.from_key(reader.read_string())
            value = reader.read_uint8_array()
        return Asset(identifier, value)


class FileArraySerializer:
    @classmethod
    def serialize(cls, item: list[Asset]) -> bytes:
        writer = ByteWriter()
        writer.write_uleb128(len(item))
        for file in item:
            writer.write_string(file.id.key())
            writer.write_uint8_array(file.buffer)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> list[Asset]:
        with ByteReader(item) as reader:
            count = reader.read_uleb128()
            files: list[Asset] = []
            for _ in range(count):
                identifier = Identifier.from_key(reader.read_string())
                value = reader.read_uint8_array()
                files.append(Asset(identifier, value))
        return files


ASSET_UPLOAD_PERMISSION_ID = ASSET_EXTENSION_TYPE / "upload"
ASSET_UPLOAD_ENDPOINT = EndpointType[Asset, Identifier].create_serialized(
    ASSET_EXTENSION_TYPE,
    "upload",
    request_serializer=FileSerializer,
    response_serializer=Serializer.model(Identifier).to_json(),
    permission_id=ASSET_UPLOAD_PERMISSION_ID,
)
ASSET_UPLOAD_MANY_ENDPOINT = EndpointType[list[Asset], list[Identifier]].create_serialized(
    ASSET_EXTENSION_TYPE,
    "upload_many",
    request_serializer=FileArraySerializer,
    response_serializer=Serializer.model(Identifier).to_array().to_json(),
    permission_id=ASSET_UPLOAD_PERMISSION_ID,
)
ASSET_DOWNLOAD_PERMISSION_ID = ASSET_EXTENSION_TYPE / "download"
ASSET_DOWNLOAD_ENDPOINT = EndpointType[Identifier, Asset].create_serialized(
    ASSET_EXTENSION_TYPE,
    "download",
    request_serializer=Serializer.model(Identifier).to_json(),
    response_serializer=FileSerializer,
    permission_id=ASSET_DOWNLOAD_PERMISSION_ID,
)
ASSET_DOWNLOAD_MANY_ENDPOINT = EndpointType[list[Identifier], list[Asset]].create_serialized(
    ASSET_EXTENSION_TYPE,
    "download_many",
    request_serializer=Serializer.model(Identifier).to_array().to_json(),
    response_serializer=FileArraySerializer,
    permission_id=ASSET_DOWNLOAD_PERMISSION_ID,
)
ASSET_DELETE_PERMISSION_ID = ASSET_EXTENSION_TYPE / "delete"
ASSET_DELETE_ENDPOINT = EndpointType[Identifier, None].create_serialized(
    ASSET_EXTENSION_TYPE,
    "delete",
    request_serializer=Serializer.model(Identifier).to_json(),
    response_serializer=Serializer.json(),
    permission_id=ASSET_DELETE_PERMISSION_ID,
)


class AssetExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return ASSET_EXTENSION_TYPE

    def __init__(self, omu: Omu) -> None:
        self.omu = omu

    async def upload(self, file: Asset) -> Identifier:
        return await self.omu.endpoints.call(ASSET_UPLOAD_ENDPOINT, file)

    async def upload_many(self, files: list[Asset]) -> list[Identifier]:
        return await self.omu.endpoints.call(ASSET_UPLOAD_MANY_ENDPOINT, files)

    async def download(self, identifier: Identifier) -> Asset:
        return await self.omu.endpoints.call(ASSET_DOWNLOAD_ENDPOINT, identifier)

    async def download_many(self, identifiers: list[Identifier]) -> list[Asset]:
        return await self.omu.endpoints.call(ASSET_DOWNLOAD_MANY_ENDPOINT, identifiers)

    async def delete(self, identifier: Identifier) -> None:
        await self.omu.endpoints.call(ASSET_DELETE_ENDPOINT, identifier)

    def url(self, identifier: Identifier) -> str:
        address = self.omu.network.address
        protocol = "https" if address.secure else "http"
        return f"{protocol}://{address.host}:{address.port}/asset?id={identifier.key()}"

    def proxy(self, url: str) -> str:
        address = self.omu.network.address
        protocol = "https" if address.secure else "http"
        return f"{protocol}://{address.host}:{address.port}/proxy?url={url}"



from __future__ import annotations

from dataclasses import dataclass

from omu.bytebuffer import ByteReader, ByteWriter
from omu.identifier import Identifier

from .registry import RegistryPermissions


@dataclass(frozen=True, slots=True)
class RegistryPacket:
    id: Identifier
    value: bytes | None

    @classmethod
    def serialize(cls, item: RegistryPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_boolean(item.value is not None)
        if item.value is not None:
            writer.write_uint8_array(item.value)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> RegistryPacket:
        with ByteReader(item) as reader:
            key = Identifier.from_key(reader.read_string())
            existing = reader.read_boolean()
            value = reader.read_uint8_array() if existing else None
        return RegistryPacket(key, value)


@dataclass(frozen=True, slots=True)
class RegistryRegisterPacket:
    id: Identifier
    permissions: RegistryPermissions

    @classmethod
    def serialize(cls, item: RegistryRegisterPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        item.permissions.serialize(writer)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> RegistryRegisterPacket:
        with ByteReader(item) as reader:
            key = Identifier.from_key(reader.read_string())
            permissions = RegistryPermissions.deserialize(reader)
        return RegistryRegisterPacket(key, permissions)



from __future__ import annotations

import abc
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from omu.bytebuffer import ByteReader, ByteWriter, Flags
from omu.event_emitter import Unlisten
from omu.helper import Coro, map_optional
from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


@dataclass(frozen=True, slots=True)
class RegistryPermissions:
    all: Identifier | None = None
    read: Identifier | None = None
    write: Identifier | None = None

    def serialize(self, writer: ByteWriter) -> None:
        flags = Flags(0, 3)
        flags.set(0, self.all is not None)
        flags.set(1, self.read is not None)
        flags.set(2, self.write is not None)
        writer.write_flags(flags)
        if self.all is not None:
            writer.write_string(self.all.key())
        if self.read is not None:
            writer.write_string(self.read.key())
        if self.write is not None:
            writer.write_string(self.write.key())

    @classmethod
    def deserialize(cls, reader: ByteReader) -> RegistryPermissions:
        flags = reader.read_flags(3)
        all_id = reader.read_string() if flags.has(0) else None
        read_id = reader.read_string() if flags.has(1) else None
        write_id = reader.read_string() if flags.has(2) else None
        return RegistryPermissions(
            map_optional(all_id, Identifier.from_key),
            map_optional(read_id, Identifier.from_key),
            map_optional(write_id, Identifier.from_key),
        )


@dataclass(frozen=True, slots=True)
class RegistryType[T]:
    id: Identifier
    default_value: T
    serializer: Serializable[T, bytes]
    permissions: RegistryPermissions = RegistryPermissions()

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
        default_value: T,
        permissions: RegistryPermissions | None = None,
    ) -> RegistryType[T]:
        return cls(
            identifier / name,
            default_value,
            Serializer.json(),
            permissions or RegistryPermissions(),
        )

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        default_value: T,
        serializer: Serializable[T, bytes],
        permissions: RegistryPermissions | None = None,
    ) -> RegistryType[T]:
        return cls(
            identifier / name,
            default_value,
            serializer,
            permissions or RegistryPermissions(),
        )


class Registry[T](abc.ABC):
    @property
    @abc.abstractmethod
    def value(self) -> T: ...

    @abc.abstractmethod
    async def get(self) -> T: ...

    @abc.abstractmethod
    async def set(self, value: T) -> None: ...

    @abc.abstractmethod
    async def update(self, handler: Coro[[T], T] | Callable[[T], T]) -> T: ...

    @abc.abstractmethod
    async def modify(self, handler: Coro[[T], Any] | Callable[[T], Any]) -> T: ...

    @abc.abstractmethod
    def listen(self, handler: Coro[[T], None]) -> Unlisten: ...



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Literal

from loguru import logger

from omu.address import Address
from omu.app import AppType
from omu.brand import BRAND
from omu.errors import (
    AnotherConnection,
    InvalidOrigin,
    InvalidPacket,
    InvalidToken,
    InvalidVersion,
    OmuError,
    PermissionDenied,
)
from omu.event_emitter import EventEmitter
from omu.helper import Coro
from omu.identifier import Identifier
from omu.omu import Omu
from omu.token import TokenProvider

from .connection import CloseError, Connection, Transport
from .encryption import AES, Decryptor, Encryptor
from .packet import Packet, PacketType
from .packet.packet_types import (
    PACKET_TYPES,
    ConnectPacket,
    DisconnectPacket,
    DisconnectType,
    EncryptionResponse,
)
from .packet_mapper import PacketMapper


@dataclass(frozen=True, slots=True)
class PacketHandler[T]:
    packet_type: PacketType[T]
    event: EventEmitter[T]


class Network:
    def __init__(
        self,
        omu: Omu,
        address: Address,
        token_provider: TokenProvider,
        transport: Transport,
        connection: Connection | None = None,
    ):
        self.omu = omu
        self.address = address
        self._token_provider = token_provider
        self.transport = transport
        self.connection: Connection | None = connection
        self._event = NetworkEvents()
        self._tasks: list[Coro[[], None]] = []
        self._packet_mapper = PacketMapper()
        self._packet_handlers: dict[Identifier, PacketHandler] = {}
        self.listen_task: asyncio.Task[None] | None = None
        self.register_packet(
            PACKET_TYPES.SERVER_META,
            PACKET_TYPES.CONNECT,
            PACKET_TYPES.DISCONNECT,
            PACKET_TYPES.TOKEN,
            PACKET_TYPES.READY,
            PACKET_TYPES.ENCRYPTED_PACKET,
        )
        self.add_packet_handler(PACKET_TYPES.TOKEN, self.handle_token)
        self.add_packet_handler(PACKET_TYPES.DISCONNECT, self.handle_disconnect)
        self.aes: AES | None = None

    async def handle_token(self, token: str | None):
        if token is None:
            return
        self._token_provider.store(self.address, self.omu.app, token)

    async def handle_disconnect(self, reason: DisconnectPacket):
        if reason.type in {
            DisconnectType.SHUTDOWN,
            DisconnectType.CLOSE,
            DisconnectType.SERVER_RESTART,
        }:
            return

        ERROR_MAP: dict[DisconnectType, type[OmuError]] = {
            DisconnectType.ANOTHER_CONNECTION: AnotherConnection,
            DisconnectType.PERMISSION_DENIED: PermissionDenied,
            DisconnectType.INVALID_TOKEN: InvalidToken,
            DisconnectType.INVALID_ORIGIN: InvalidOrigin,
            DisconnectType.INVALID_VERSION: InvalidVersion,
            DisconnectType.INVALID_PACKET: InvalidPacket,
            DisconnectType.INVALID_PACKET_TYPE: InvalidPacket,
            DisconnectType.INVALID_PACKET_DATA: InvalidPacket,
        }
        error = ERROR_MAP.get(reason.type)
        if error:
            raise error(reason.message or reason.type.value)

    def set_connection(self, connection: Connection) -> None:
        if self.connection:
            raise RuntimeError("Cannot change connection while connected")
        if self.connection:
            del self.connection
        self.connection = connection

    def set_transport(self, transport: Transport) -> None:
        if self.connection:
            raise RuntimeError("Cannot change connection while connected")
        if self.connection:
            del self.connection
        self.transport = transport

    def set_token_provider(self, token_provider: TokenProvider) -> None:
        self._token_provider = token_provider

    def register_packet(self, *packet_types: PacketType) -> None:
        self._packet_mapper.register(*packet_types)
        for packet_type in packet_types:
            if self._packet_handlers.get(packet_type.id):
                raise ValueError(f"Packet type {packet_type.id} already registered")
            self._packet_handlers[packet_type.id] = PacketHandler(
                packet_type,
                EventEmitter(),
            )

    def add_packet_handler[T](
        self,
        packet_type: PacketType[T],
        packet_handler: Coro[[T], None] | None = None,
    ):
        if not self._packet_handlers.get(packet_type.id):
            raise ValueError(f"Packet type {packet_type.id} not registered")

        def decorator(func: Coro[[T], None]) -> None:
            self._packet_handlers[packet_type.id].event.listen(func)

        if packet_handler:
            decorator(packet_handler)
        return decorator

    @property
    def connected(self) -> bool:
        if not self.connection:
            return False
        return not self.connection.closed

    async def connect(self, *, reconnect: bool = True) -> None:
        if self.listen_task:
            raise RuntimeError("Already connected")

        exception: Exception | None = None
        attempts = 0
        while True:
            try:
                self.connection = self.connection or await self.transport.connect()

                meta = (await self.connection.receive_as(self._packet_mapper, PACKET_TYPES.SERVER_META)).unwrap()
                self.address = self.address.with_hash(meta["hash"])
                token = self._token_provider.get(self.address, self.omu.app)
                if token is None:
                    raise InvalidToken(f"No token for {self.address} and app {self.omu.app.id}")
                encryption_resp: EncryptionResponse | None = None
                if self.omu.app.type == AppType.REMOTE and meta["encryption"]:
                    decryptor = Decryptor.new()
                    encryptor = Encryptor.new(meta["encryption"]["rsa"])
                    if token:
                        token = encryptor.encrypt_string(token)
                    aes = AES.new()
                    encryption_resp = {
                        "kind": "v1",
                        "rsa": decryptor.to_request(),
                        "aes": aes.serialize(encryptor),
                    }
                    self.aes = aes
                await self.connection.send(
                    Packet(
                        PACKET_TYPES.CONNECT,
                        ConnectPacket(
                            app=self.omu.app,
                            protocol={"version": self.omu.version, "brand": BRAND},
                            encryption=encryption_resp,
                            token=token,
                        ),
                    ),
                    self._packet_mapper,
                )
                self.listen_task = asyncio.create_task(self._listen_task())
                await self._event.status.emit("connected")
                await self._event.connected.emit()
                await self._dispatch_tasks()

                await self.send(Packet(PACKET_TYPES.READY, None))
                await self.listen_task
            except Exception as e:
                if reconnect:
                    attempts += 1
                    if attempts > 10:
                        return
                    if exception:
                        logger.error("Failed to reconnect")
                    else:
                        logger.opt(exception=e).error(f"Failed to connect to {self.address.host}:{self.address.port}")
                    exception = e
                    continue
                else:
                    raise e
            finally:
                self.listen_task = None

            if not reconnect:
                break

            await asyncio.sleep(1)

    async def disconnect(self) -> None:
        if not self.connection:
            return
        await self.send(
            Packet(
                PACKET_TYPES.DISCONNECT,
                DisconnectPacket(DisconnectType.CLOSE, "Client disconnected"),
            )
        )
        await self.connection.close()
        self.connection = None
        await self._event.status.emit("disconnected")
        await self._event.disconnected.emit()

    async def close(self) -> None:
        await self.disconnect()

    async def send(self, packet: Packet) -> None:
        if not self.connection:
            raise RuntimeError("Not connected")
        if self.aes:
            serialized = self._packet_mapper.serialize(packet)
            packet = self.aes.encrypt(serialized)
        await self.connection.send(packet, self._packet_mapper)

    async def _listen_task(self):
        try:
            while self.connection:
                received = await self.connection.receive(self._packet_mapper)
                if received.is_err is True:
                    logger.error(received.err.message)
                    return
                asyncio.create_task(self.dispatch_packet(received.value))
        except CloseError as e:
            logger.opt(exception=e).error("Connection closed")
        finally:
            await self.disconnect()

    async def dispatch_packet(self, packet: Packet) -> None:
        if PACKET_TYPES.ENCRYPTED_PACKET.match(packet):
            if self.aes is None:
                raise RuntimeError("Received encrypted packet before encryption was established")
            decrypted = self.aes.decrypt(packet)
            packet = self._packet_mapper.deserialize(decrypted)
        await self._event.packet.emit(packet)
        packet_handler = self._packet_handlers.get(packet.type.id)
        if not packet_handler:
            return
        await packet_handler.event(packet.data)

    @property
    def event(self) -> NetworkEvents:
        return self._event

    def add_task(self, task: Coro[[], None]) -> None:
        if self.omu.ready:
            raise RuntimeError("Cannot add task after client is ready")
        self._tasks.append(task)

    def remove_task(self, task: Coro[[], None]) -> None:
        self._tasks.remove(task)

    async def _dispatch_tasks(self) -> None:
        for task in self._tasks:
            await task()


type NetworkStatus = Literal["connecting", "connected", "disconnected"]


class NetworkEvents:
    def __init__(self) -> None:
        self.connected = EventEmitter[[]]()
        self.disconnected = EventEmitter[[]]()
        self.packet = EventEmitter[Packet]()
        self.status = EventEmitter[NetworkStatus]()



from __future__ import annotations

import json
from dataclasses import dataclass

from omu.bytebuffer import ByteReader, ByteWriter
from omu.identifier import Identifier


@dataclass(frozen=True, slots=True)
class EndpointRegisterPacket:
    endpoints: dict[Identifier, Identifier | None]

    @classmethod
    def serialize(cls, item: EndpointRegisterPacket) -> bytes:
        writer = ByteWriter()
        writer.write_uleb128(len(item.endpoints))
        for key, value in item.endpoints.items():
            writer.write_string(key.key())
            writer.write_string(value.key() if value else "")
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> EndpointRegisterPacket:
        with ByteReader(item) as reader:
            count = reader.read_uleb128()
            endpoints = {}
            for _ in range(count):
                key = reader.read_string()
                value = reader.read_string()
                endpoints[Identifier.from_key(key)] = Identifier.from_key(value) if value else None
        return EndpointRegisterPacket(endpoints=endpoints)


@dataclass(frozen=True, slots=True)
class InvokedParams:
    id: Identifier
    caller: Identifier
    key: int

    @staticmethod
    def serialize(item: InvokedParams) -> str:
        return json.dumps(
            {
                "id": item.id.key(),
                "caller": item.caller.key(),
                "key": item.key,
            }
        )

    @staticmethod
    def deserialize(item: str) -> InvokedParams:
        data = json.loads(item)
        assert isinstance(data["id"], str)
        assert isinstance(data["caller"], str)
        assert isinstance(data["key"], int)
        return InvokedParams(
            id=Identifier.from_key(data["id"]),
            caller=Identifier.from_key(data["caller"]),
            key=data["key"],
        )


@dataclass(frozen=True, slots=True)
class EndpointInvokedPacket:
    params: InvokedParams
    buffer: bytes

    @classmethod
    def serialize(cls, item: EndpointInvokedPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(InvokedParams.serialize(item.params))
        writer.write_uint8_array(item.buffer)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> EndpointInvokedPacket:
        with ByteReader(item) as reader:
            params = InvokedParams.deserialize(reader.read_string())
            buffer = reader.read_uint8_array()
        return EndpointInvokedPacket(
            params=params,
            buffer=buffer,
        )


@dataclass(frozen=True, slots=True)
class InvokeParams:
    id: Identifier
    key: int

    @staticmethod
    def serialize(item: InvokeParams) -> str:
        return json.dumps(
            {
                "id": item.id.key(),
                "key": item.key,
            }
        )

    @staticmethod
    def deserialize(item: str) -> InvokeParams:
        data = json.loads(item)
        assert isinstance(data["id"], str)
        assert isinstance(data["key"], int)
        return InvokeParams(
            id=Identifier.from_key(data["id"]),
            key=data["key"],
        )


@dataclass(frozen=True, slots=True)
class EndpointInvokePacket:
    params: InvokeParams
    buffer: bytes

    @classmethod
    def serialize(cls, item: EndpointInvokePacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(InvokeParams.serialize(item.params))
        writer.write_uint8_array(item.buffer)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> EndpointInvokePacket:
        with ByteReader(item) as reader:
            params = InvokeParams.deserialize(reader.read_string())
            buffer = reader.read_uint8_array()
        return EndpointInvokePacket(
            params=params,
            buffer=buffer,
        )


@dataclass(frozen=True, slots=True)
class ResponseParams:
    id: Identifier
    key: int
    error: str | None

    @staticmethod
    def serialize(item: ResponseParams) -> str:
        return json.dumps(
            {
                "id": item.id.key(),
                "key": item.key,
                "error": item.error,
            }
        )

    @staticmethod
    def deserialize(item: str) -> ResponseParams:
        data = json.loads(item)
        assert isinstance(data["id"], str)
        assert isinstance(data["key"], int)
        assert isinstance(data["error"], str | None)
        return ResponseParams(
            id=Identifier.from_key(data["id"]),
            key=data["key"],
            error=data["error"],
        )


@dataclass(frozen=True, slots=True)
class EndpointResponsePacket:
    params: ResponseParams
    buffer: bytes

    @classmethod
    def serialize(cls, item: EndpointResponsePacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(ResponseParams.serialize(item.params))
        writer.write_uint8_array(item.buffer)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> EndpointResponsePacket:
        with ByteReader(item) as reader:
            params = ResponseParams.deserialize(reader.read_string())
            buffer = reader.read_uint8_array()
        return EndpointResponsePacket(
            params=params,
            buffer=buffer,
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Literal, TypedDict

from omu.app import App, AppJson
from omu.identifier import Identifier
from omu.serializer import Serializer

from .packet import PacketType


class ProtocolInfo(TypedDict):
    brand: str
    version: str


class RSANumbers(TypedDict):
    e: str
    n: str


class EncryptionRequest(TypedDict):
    kind: Literal["v1"]
    rsa: RSANumbers


class EncryptionResponse(TypedDict):
    kind: Literal["v1"]
    rsa: RSANumbers
    aes: str


class ServerMeta(TypedDict):
    protocol: ProtocolInfo
    encryption: EncryptionRequest | None
    hash: str | None


class ConnectPacketData(TypedDict):
    protocol: ProtocolInfo
    app: AppJson
    token: str
    encryption: EncryptionResponse | None


@dataclass(frozen=True, slots=True)
class ConnectPacket:
    app: App
    protocol: ProtocolInfo
    token: str
    encryption: EncryptionResponse | None = None

    @staticmethod
    def serialize(packet: ConnectPacket) -> ConnectPacketData:
        return {
            "app": packet.app.to_json(),
            "protocol": packet.protocol,
            "encryption": packet.encryption,
            "token": packet.token,
        }

    @staticmethod
    def deserialize(json: ConnectPacketData) -> ConnectPacket:
        return ConnectPacket(
            app=App.from_json(json["app"]),
            protocol=json["protocol"],
            token=json["token"],
            encryption=json["encryption"],
        )


class DisconnectType(str, Enum):
    INVALID_TOKEN = "invalid_token"
    INVALID_ORIGIN = "invalid_origin"
    INVALID_VERSION = "invalid_version"
    INVALID_PACKET_TYPE = "invalid_packet_type"
    INVALID_PACKET_DATA = "invalid_packet_data"
    INVALID_PACKET = "invalid_packet"
    INTERNAL_ERROR = "internal_error"
    ANOTHER_CONNECTION = "another_connection"
    PERMISSION_DENIED = "permission_denied"
    SERVER_RESTART = "server_restart"
    SHUTDOWN = "shutdown"
    CLOSE = "close"


class DisconnectPacketData(TypedDict):
    type: str
    message: str | None


class DisconnectPacket:
    def __init__(self, type: DisconnectType, message: str | None = None):
        self.type: DisconnectType = type
        self.message = message

    @staticmethod
    def serialize(packet: DisconnectPacket) -> DisconnectPacketData:
        return {
            "type": packet.type.value,
            "message": packet.message,
        }

    @staticmethod
    def deserialize(json: DisconnectPacketData) -> DisconnectPacket:
        return DisconnectPacket(
            type=DisconnectType(json["type"]),
            message=json["message"],
        )


CORE_ID = Identifier("core", "packet")


class Authenticate(TypedDict):
    kind: Literal["v1"]
    token: str


"""
[C->S] connect to server
[C<-S] rsa public key sharing and send metadata to authenticate if target is a valid server.
[C->S] rsa public key sharing and send AES CBC key
// Encryption starts here and all packets are encrypted with AES CBC using the shared key
"""


class PACKET_TYPES:
    SERVER_META = PacketType[ServerMeta].create_json(
        CORE_ID,
        "server_meta",
    )
    CONNECT = PacketType[ConnectPacket].create_json(
        CORE_ID,
        "connect",
        ConnectPacket,
    )
    DISCONNECT = PacketType[DisconnectPacket].create_json(
        CORE_ID,
        "disconnect",
        DisconnectPacket,
    )
    TOKEN = PacketType[str | None].create_json(
        CORE_ID,
        "token",
    )
    READY = PacketType[None].create_json(
        CORE_ID,
        "ready",
    )
    ENCRYPTED_PACKET = PacketType[bytes].create_serialized(
        CORE_ID,
        name="e",
        serializer=Serializer.noop(),
    )



from __future__ import annotations

import abc
from collections.abc import Callable
from typing import TYPE_CHECKING

from omu.identifier import Identifier

if TYPE_CHECKING:
    from omu.omu import Omu


class Extension(abc.ABC):
    @property
    @abc.abstractmethod
    def type(self) -> ExtensionType: ...


EXT_NAMESPACE = "ext"


class ExtensionType[T: Extension](Identifier):
    name: str
    create: Callable[[Omu], T]

    def __init__(
        self,
        name: str,
        create: Callable[[Omu], T],
    ) -> None:
        super().__init__(EXT_NAMESPACE, name)
        self.name = name
        self.create = create

    def key(self) -> str:
        return self.name



from __future__ import annotations

from dataclasses import dataclass

from omu.bytebuffer import ByteReader, ByteWriter


@dataclass(frozen=True, slots=True)
class ConsolePacket:
    lines: list[str]

    @classmethod
    def serialize(cls, item: ConsolePacket) -> bytes:
        writer = ByteWriter()
        writer.write_uleb128(len(item.lines))
        for line in item.lines:
            writer.write_string(line)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> ConsolePacket:
        with ByteReader(item) as reader:
            line_count = reader.read_uleb128()
            lines = [reader.read_string() for _ in range(line_count)]
        return ConsolePacket(lines)



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Any

from omu.bytebuffer import ByteReader, ByteWriter, Flags
from omu.event_emitter import Unlisten
from omu.helper import Coro
from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


@dataclass(frozen=True, slots=True)
class SignalPermissions:
    all: Identifier | None = None
    listen: Identifier | None = None
    notify: Identifier | None = None

    def serialize(self, writer: ByteWriter) -> None:
        flags = Flags(length=3)
        flags.set(0, self.all is not None)
        flags.set(1, self.listen is not None)
        flags.set(2, self.notify is not None)
        writer.write_uint8(flags.value)
        if self.all is not None:
            writer.write_string(self.all.key())
        if self.listen is not None:
            writer.write_string(self.listen.key())
        if self.notify is not None:
            writer.write_string(self.notify.key())

    @classmethod
    def deserialize(cls, reader: ByteReader) -> SignalPermissions:
        flags = Flags(reader.read_uint8())
        all = flags.if_set(0, lambda: Identifier.from_key(reader.read_string()))
        listen = flags.if_set(1, lambda: Identifier.from_key(reader.read_string()))
        send = flags.if_set(2, lambda: Identifier.from_key(reader.read_string()))
        return SignalPermissions(all=all, listen=listen, notify=send)


@dataclass(frozen=True, slots=True)
class SignalType[T]:
    id: Identifier
    serializer: Serializable[T, bytes]
    permissions: SignalPermissions = SignalPermissions()

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[T, Any] | None = None,
        permissions: SignalPermissions | None = None,
    ):
        return cls(
            id=identifier / name,
            serializer=Serializer.of(serializer or Serializer.noop()).to_json(),
            permissions=permissions or SignalPermissions(),
        )

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[T, bytes],
        permissions: SignalPermissions | None = None,
    ):
        return cls(
            id=identifier / name,
            serializer=serializer,
            permissions=permissions or SignalPermissions(),
        )


class Signal[T](abc.ABC):
    @abc.abstractmethod
    def listen(self, listener: Coro[[T], None]) -> Unlisten: ...

    @abc.abstractmethod
    async def notify(self, body: T) -> None: ...



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol, TypeGuard

from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


@dataclass(frozen=True, slots=True)
class PacketData:
    type: str
    data: bytes


@dataclass(frozen=True, slots=True)
class Packet[T]:
    type: PacketType[T]
    data: T


class PacketClass[T](Protocol):
    def serialize(self, item: T) -> bytes: ...

    def deserialize(self, item: bytes) -> T: ...


@dataclass(frozen=True, slots=True)
class PacketType[T]:
    id: Identifier
    serializer: Serializable[T, bytes]

    def new(self, data: T) -> Packet[T]:
        return Packet(self, data)

    def match(self, packet: Packet) -> TypeGuard[Packet[T]]:
        return packet.type == self

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[T, Any] | None = None,
    ) -> PacketType[T]:
        return PacketType(
            id=identifier / name,
            serializer=Serializer.of(serializer or Serializer.noop()).to_json(),
        )

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[T, bytes],
    ) -> PacketType[T]:
        return PacketType(
            id=identifier / name,
            serializer=serializer,
        )

    @classmethod
    def create(
        cls,
        identifier: Identifier,
        name: str,
        type_class: PacketClass[T],
    ) -> PacketType[T]:
        return PacketType(
            id=identifier / name,
            serializer=type_class,
        )
